// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "avatar.h"

// vektor totozny s osou Y
vec3 Vup(0.0f,1.0f,0.0f);
// vektor totozny s osou -Z
vec3 Vforward(0.0f,0.0f,-1.0f);

// konstruktor *************************************************************
CAvatar::CAvatar()
{
 reset(); // reset parametrov Avatara

 // scitanie koeficientov kvadratickych funkcii
 // pocet koeficientov - pocet vzoriek / s = FPS - v intervale <0;MAXFPS>
 // a - predstavuje cas, ktory sa priratava pri skoku/posobeni gravitacie
 // tento cas je zavisly na FPS, Avatar pri roznych FPS prekona vzdy
 // rovnaku vzdialenost - skace rovnako vysoko a pada rovnako hlboko
 // K - konstanta rovnaka pre akekolvek FPS = vyska skoku
 //
 // 0a^2 + 1a^2 + 4a^2 + 9a^2 + 16a^2 + 25a^2 + 36a^2 + ... = K
 // sucet koeficientov x a^2 = K
 //                      a^2 = K / sucet koeficientov
 //                        a = sqrt ( K / sucet koeficientov )
 //
 // pouzity pristup k vypoctu priebehu skoku/padu zavisleho na FPS
 // je matematicky spravny, rozdiely vo vyske skoku vznikaju pri nahlych
 // zmenach FPS, co vsak nemozno nijako matematicky ovplyvnit,
 // cas pri skoku/pade plynie rychlejsie ako FPS
 loopk(MAXFPS+1) {
  gKoef[k]=0.0f; loopi(k+1) { gKoef[k]+=(GLfloat) (i*i); }
  // INFO: sucet koeficientov
  //printf("%u %f\n",k,gKoef[k]);
  if (k==0) continue;
  // prirastok casu pre skok/pad - zavisly na FPS
  gKoef[k]=sqrtf(2.2f/gKoef[k]);
 }
}

// destruktor **************************************************************
CAvatar::~CAvatar()
{

}

// inicializacia Avatara ***************************************************
void CAvatar::init(void)
{
 glMatrixMode(GL_PROJECTION); // vyber projekcnej matice
 glLoadIdentity();            // zresetovanie projekcnej matice

 if (scr_h==0) scr_h=1;
 // nastavenie perspektivneho premietania
 // fov - zorny uhol pohladu, W/H - aspect - pomer sirky a vysky kresliacej plochy
 // viewnear - minimalny dohlad, viewfar - maximalny dohlad
 gluPerspective(fov,(GLfloat) scr_w/(GLfloat) scr_h,viewnear,viewfar);

 // nastavenie rozmerov OpenGL kresliacej plochy
 glViewport(0,0,scr_w,scr_h);

 // TRANSFORMUJEME SCENU, NIE KAMERU!
 glMatrixMode(GL_MODELVIEW); // vyber modelview matice
 glLoadIdentity();           // zresetovanie modelview matice
}

// reset parametrov Avatara ************************************************
void CAvatar::reset(void)
{
 // sposob vypoctu pohybu, 0=chodza, 1=let
 movetype=0;

 avatarEye=1.4f;   // vyska oci Avatara
 avatarWidth=0.8f; // polovica sirky Avatara
 avatarSlope=0.5f; // sklon roviny, ktory dokaze Avatar prekonat

 mat.identity(); // nastavenie jednotkovej transformacnej matice Avatara

 actpos.set(-23.0f,-avatarEye+6.6f,-19.0f); oldpos=actpos; // aktualna a povodna poloha
 //actpos.set(0.0f,-avatarEye+2.0f,0.0f); oldpos=actpos; // aktualna a povodna poloha
 eyepos.set(0.0f,0.0f,0.0f);                               // pociatocna poloha oci
 targetV.set(0.0f,0.0f,0.0f);                      // vektor smerujuci dopredu
 move.set(0.0f,0.0f,0.0f);                         // prirastok pre zmenu polohy
 oldmove.set(0.0f,0.0f,0.0f);                      // prirastok pre zmenu polohy
 angle.set(0.0f,0.0f,0.0f);                        // natocenie
 rotate.set(0.0f,0.0f,0.0f);                       // prirastok pre zmenu natocenia

 tilt=0.0f; tiltAbs=0.0f; // kracanie Avatara
 goTilt=false;
 lean=0.0f; leanAdd=0.0f; // uklon Avatara od zvislej roviny

 canRotate=1.0f; // Avatar sa moze natacat

 // koeficienty
 speed=2.0f; slide=2.0f; altit=2.0f; // rychlost, bocny pohyb, zmena vysky
 gravity=3.0f;                       // gravitacne zrychlenie
 viewnear=0.1f; viewfar=1000.0f;     // minimalny a maximalny dohlad
 fov=60.0f;                          // zorny uhol pohladu
 jumpSpeed=50;                      // rychlost skoku
 jumpTime=0.0f;                      // cas zacatia skoku

 // akceleracia Avatara
 accelX=0.0f; accelZ=0.0f;
 accelAddX=0.0f; accelAddZ=0.0f;

 // aktualny cas posobenia gravitacie
 gravityTime=0.0f;

 // Avatar stoji
 crouch=1.0f;

 // indikatory pre vykonavanie definovanych posunov Avatara
 goForward=0.0f; goBack=0.0f; goUp=0.0f; goDown=0.0f; goRight=0.0f; goLeft=0.0f;
 goLeanLeft=0.0f; goLeanRight=0.0f;
 goCrouch=0.0f;
 goJump=0.0f; goFast=0.0f;
 // strelba
 isFire1=0; isFire2=0;

 // indikatory detekovanych kolizii
 isWallCollision=false; isFloorCollision=false; isRoofCollision=false;
}

// nastavenie prirastkov pre zmenu polohy Avatara **************************
/*
 rychlost  v=s/t [m/s]  draha  s=vt [m]
           v=at [m/s]
 zrychlenie             a=s/t2 [m/s2]
 gravitacne zrychlenie  g=v/t [m/s2], g na Zemi=9.81 m/s2
                        g=s/t2  draha  s=gt2 [m]
*/
void CAvatar::setmovxyz(void)
{
 // ulozenie povodneho prirastku pre zmenu polohy
 oldmove=move;

 // ak su stlacene sucasne dve protichodne klavesy, Avatar sa nehybe
 // niektore hodnoty su upravene pre dosiahnutie realnejsieho vnemu
 goFast=2.0f+8.0f*goFast;

 // Avatar nekraca
 goTilt=false;

 switch (movetype) {
  case 0:  // chodza
   rightLeft=(goRight-goLeft);     // tlacidla vlavo/vpravo
   forwardBack=(goForward-goBack); // tlacidla dopredu/dozadu

   move.set(-goFast*rightLeft*slide,    // vlavo/vpravo
             goFast*gravity,            // gravitacia
            -goFast*forwardBack*speed); // dopredu/dozadu

   // uprava pohybu doprava/dolava
   if (rightLeft!=0.0f) {
    accelAddX=1.5f; // Avatar zrychluje v smere X
    tilt=-move.x*1.5f; goTilt=true; // Avatar zacina kracat
   }
   else { // Avatar spomaluje
    move.x=oldmove.x; accelAddX=-1.5f;
   }

   // uprava pohybu dopredu/dozadu
   if (forwardBack!=0.0f) {
    accelAddZ=1.5f; // Avatar zrychluje v smere Z
    tilt=-move.z*1.5f; goTilt=true; // Avatar zacina kracat
   }
   else { // Avatar spomaluje
    move.z=oldmove.z; accelAddZ=-1.5f;
   }

   // ak Avatar zacal kracat
   if (goTilt) { tiltAbs=tilt; if (tiltAbs<0.0f) tiltAbs=-tiltAbs; }
   else { // inak ukoncenie kracania
    if (rotate.z>0.0f) tilt=-tiltAbs;     // vyrovnanie dolava
    else if (rotate.z<0.0f) tilt=tiltAbs; // vyrovnanie doprava
         else tilt=0.0f;
   }

   // uklon Avatara
   leanAdd=(goLeanRight-goLeanLeft);
   if (leanAdd==0.0f) {
    if (lean>0.0f) leanAdd=-0.5f;
    else if (lean<0.0f) leanAdd=0.5f;
   }

   // ak je Avatar na podlahe, moze skakat
   // rychlost skoku = hodnota FPS, ktora sa pouzije na vyber prirastkov casu
   // cas zacatia skoku = FPS, o ktore sa posunie cas posobenia gravitacie
   if (jumpTime==0.0f) {
    if (difTime!=0.0f) jumpSpeed=(GLuint) (1.0f/(difTime*2.0f));
    jumpTime=goJump*((GLfloat) jumpSpeed);
    // INFO: rychlost a cas zacatia skoku
    //printf("jumpSpeed: %u jumpTime: %f\n",jumpSpeed,jumpTime);
   }

   if (goCrouch==1.0f) crouchAdd=-5.0f; // Avatar si cupne
   else {                               // Avatar vstava
    if (crouchAdd!=5.0f) {
     crouch=1.0f; detectRoofCollision(); crouch=0.65f;
     if (!isRoofCollision) crouchAdd=5.0f;
    }
   }
 break;
  case 1:  // let
   move.set(-goFast*(goRight-goLeft)*slide,    // vlavo a vpravo
            -goFast*(goUp-goDown)*altit,       // hore a dole
            -goFast*(goForward-goBack)*speed); // dopredu a dozadu
  break;
  default: // dalsie typy pohybu...
  break;
 }

 // zrusenie dalsieho mozneho posunu, alebo skoku
 goForward=0.0f; goBack=0.0f; goUp=0.0f; goDown=0.0f; goRight=0.0f; goLeft=0.0f;
 goLeanLeft=0.0f; goLeanRight=0.0f;
 goJump=0.0f; goFast=0.0f;
}

// nastavenie prirastkov pre zmenu natocenia Avatara ***********************
void CAvatar::setrotxy(GLfloat X,GLfloat Y)
{
 // hodnoty su upravene pre dosiahnutie realnejsieho vnemu
 rotate.x=canRotate*X/4.0f;
 rotate.y=canRotate*Y/4.0f;

 switch (movetype) {
  case 0:  // chodza
   d1=angle.x+rotate.x;
   if ((d1<-90.0f)||(d1>90.0f)) rotate.x=0.0f;
  break;
  case 1:  // let
  break;
  default: // dalsie typy pohybu...
  break;
 }

 canRotate=1.0f; // Avatar sa nemoze natacat

 // INFO: natocenie Avatara
 //printf("%f %f %f\n",angle.x,angle.y,angle.z);
}

// nastavenie prirastkov pre zmenu natocenia Avatara ***********************
void CAvatar::setrotzl(GLfloat Z)
{
 switch (movetype) {
  case 0:  // chodza
  break;
  case 1:  // let
   rotate.z=-Z;
  break;
  default: // dalsie typy pohybu...
  break;
 }
}

// nastavenie prirastkov pre zmenu natocenia Avatara ***********************
void CAvatar::setrotzr(GLfloat Z)
{
 switch (movetype) {
  case 0:  // chodza
  break;
  case 1:  // let
   rotate.z+=Z;
  break;
  default: // dalsie typy pohybu...
  break;
 }
}

// aktualizacia parametrov Avatara *****************************************
void CAvatar::update(void)
{
 // ulozenie povodnej polohy
 oldpos=actpos;

 // SPECIALNY PRISTUP K ROTACII A NASLEDNEMU POSUNU KAMERY!
 // aktualna je vzdy modelview matica, ak zanedbame mensie nepresnosti,
 // ktore nastanu z dovodu nepouzitia glLoadIdentity(), nemusime vzdy
 // resetovat modelview maticu, ani ju nasobit transformacnou maticou Avatara

 //glMatrixMode(GL_MODELVIEW); // aktualna je vzdy modelview matica...

 // modelview matica sa nahradi transformacnou maticou Avatara
 // hodnoty prvkov pre rotaciu su inkrementovane
 // hodnoty prvkov pre posun su ponechane na 0.0f
 glLoadMatrixf(mat.get());

 switch (movetype) {
  case 0:  // chodza
   glRotatef(rotate.x,mat.m[0],mat.m[4],mat.m[8]);
   glRotatef(rotate.y,0.0f,1.0f,0.0f);
   // transformacna matica Avatara sa nahradi modelview maticou
   glGetFloatv(GL_MODELVIEW_MATRIX,mat.get());
   glLoadIdentity(); // modelview matica sa nemeni

   // vypocet vektora pre pohyb dopredu a dozadu
   // vektor je kolmy na vektory mat[0][4][8] a (0,1,0)
   p.set(mat.m[0],mat.m[4],mat.m[8]);
   Vforward=p.cross(Vup);

   // akceleracia/spomalovanie Avatara v smere X a Z - zavisle na FPS
   accelX+=accelAddX*difTime; accelZ+=accelAddZ*difTime;
   accelX=max(0.0f,min(accelX,1.0f));
   accelZ=max(0.0f,min(accelZ,1.0f));

   // vypocet posunu Avatara - zavisly na FPS
   d1=move.x*difTime*accelX; d2=move.z*difTime*accelZ;

   if (ladders.length()>0) { // v scene su definovane rebriky
    loopvk(ladders) { // pre vsetky rebriky
     if (!ladders[k].at_ladder(actpos)) { // Avatar nie je v blizkosti rebrika
      p.set(mat.m[0]*d1-Vforward.x*d2,
            move.y*difTime,
            mat.m[8]*d1-Vforward.z*d2);
      // pri chodzi posobi na Avatara gravitacia
      // inkrementacia casu posobenia gravitacie
      gravityTime+=1.0f;
      // prirastok casu pri skoku/posobeni gravitacie - zavisly na FPS
      gT=(gravityTime-jumpTime)*gKoef[jumpSpeed];

      // priebeh skoku/posobenia gravitacie, ide o standardnu kvadraticku krivku
      if (jumpTime==0.0f) { p.y+=gT*gT; }        // Avatar neskace
      else { p.y=gT*gT; if (gT<0.0f) p.y=-p.y; } // Avatar skace
     }
     else {                                // Avatar je v blizkosti rebrika
      // INFO: printf("rebrik\n");
      p.set(mat.m[0]*d1-Vforward.x*d2,
            d2,
            mat.m[8]*d1-Vforward.z*d2);
      jumpTime=0.0f; gravityTime=0.0f;
     }
    } // pre vsetky rebriky
   }
   else {                    // v scene nie su definovane rebriky
    p.set(mat.m[0]*d1-Vforward.x*d2,
          move.y*difTime,
          mat.m[8]*d1-Vforward.z*d2);
    // pri chodzi posobi na Avatara gravitacia
    // inkrementacia casu posobenia gravitacie
    gravityTime+=1.0f;
    // prirastok casu pri skoku/posobeni gravitacie - zavisly na FPS
    gT=(gravityTime-jumpTime)*gKoef[jumpSpeed];

    // priebeh skoku/posobenia gravitacie, ide o standardnu kvadraticku krivku
    if (jumpTime==0.0f) { p.y+=gT*gT; }        // Avatar neskace
    else { p.y=gT*gT; if (gT<0.0f) p.y=-p.y; } // Avatar skace
   }

   // Avatar sa moze posunut max. o polovicu svojej sirky,
   // vyriesi sa tak mnozstvo problemovych kolizii
   p.x=min(p.x,avatarWidth-0.01f);
   p.y=min(p.y,avatarWidth-0.01f);
   p.z=min(p.z,avatarWidth-0.01f);

   // nastavenie novej polohy
   actpos+=p;

   // detekcia kolizie
   detectCollision();
   // ak Avatar dopadol na podlahu, resp. narazil do stropu,
   // resetuje sa cas posobenia gravitacie a Avatar moze znovu skakat
   if ((isFloorCollision)||(isRoofCollision)) { jumpSpeed=50; jumpTime=0.0f; gravityTime=0.0f; }

   // Avatar si cupne/vstane
   crouch+=crouchAdd*difTime;
   crouch=max(0.65f,min(crouch,1.0f));
  break;
  case 1:  // let
   glRotatef(rotate.x,mat.m[0],mat.m[4],mat.m[8]);
   glRotatef(rotate.y,mat.m[1],mat.m[5],mat.m[9]);
   glRotatef(rotate.z,mat.m[2],mat.m[6],mat.m[10]);
   // transformacna matica Avatara sa nahradi modelview maticou
   glGetFloatv(GL_MODELVIEW_MATRIX,mat.get());

   // vypocet posunu Avatara - zavisly na FPS
   d1=move.x*difTime; d2=move.y*difTime; d3=move.z*difTime;
   p.set(mat.m[0]*d1+mat.m[1]*d2-mat.m[2]*d3,
         mat.m[4]*d1+mat.m[5]*d2-mat.m[6]*d3,
         mat.m[8]*d1+mat.m[9]*d2-mat.m[10]*d3);

   // nastavenie novej polohy
   actpos+=p;
  break;
  default: // dalsie typy pohybu...
  break;
 }

 // INFO: dlzka vektora stara poloha - nova poloha Avatara
 //printf("%f\n",(actpos-oldpos).magnitude());

 // nastavenie noveho natocenia
 angle+=rotate;
 // zrusenie dalsej moznej rotacie
 rotate.x=0.0f; rotate.y=0.0f; // rotate.z=0.0f;

 canRotate=1.0f; // Avatar sa moze natacat

/*
 ---------------------------------------------------------------------------
 cela tato cast je nahradena pomocou matice avamat, ktora synchronizuje
 pohyb Avatara s pohybom sceny, kracanie uz nie je simulovane matematickym
 vypoctom, ale je modelovane pomocou MD3 animacii

 // vektor smerujuci dopredu
 targetV.set(mat.m[2],mat.m[6],mat.m[10]);
 // vektor smerujuci doprava
 rightV.set(mat.m[0],mat.m[4],mat.m[8]);

 // zresetovanie modelview matice
 //glLoadIdentity(); // nemusi sa vzdy resetovat...

 // TRANSFORMACIA SCENY PODLA NASTAVENI AKTUALNEJ KAMERY!
 // ROTACIA
 //glMultMatrixf(mat.get()); // nemusi sa vzdy nasobit...
 // POSUN
 // ovplyvni modelview maticu, nie transformacnu maticu Avatara

 // chodza

 // posun sceny na poziciu oci Avatara
 glTranslatef(0.0f,-(avatarWidth+crouch*avatarEye)+rotate.z*rotate.z/60.0f,0.0f);

 // ak Avatar neskace
 if (jumpTime==0.0f) {
  // kracanie Avatara - zavisle na FPS
  d1=rotate.z;
  rotate.z+=tilt*difTime;
  if (!goTilt) { // Avatar zostane vo zvislej polohe
   if (d1*rotate.z<=0.0f) { rotate.z=0.0f; tilt=0.0f; }
  }
  if (rotate.z<-1.0f) tilt=tiltAbs; // cyklicke kracanie
  else if (rotate.z>1.0f) tilt=-tiltAbs;

  // uklon Avatara
  lean+=leanAdd;
  if (lean==0.0f) leanAdd=0.0f;
  lean=max(-9.0f,min(lean,9.0f));
 }
 else { rotate.z=0.0f; lean=0.0f; }

 // rotacia sceny okolo vektora pre pohyb dopredu a dozadu
 glRotatef(rotate.z+lean,Vforward.x,Vforward.y,Vforward.z);

 // umiestnenie sceny na poziciu pod Avatara
 glTranslatef(actpos.x,(actpos.y+avatarWidth),actpos.z);

 eyepos=actpos;
 eyepos.y-=avatarEye;
 ---------------------------------------------------------------------------
*/
 // dorobit - strelba ak je avatar nakloneny! - pomocou avamat
 //vec3 pV=eyepos-actpos;
 //pV.rotate(-DEG2RAD(lean*1.2f),targetV);
 //eyepos=actpos+pV;

 // let

 //glTranslatef(actpos.x,actpos.y,actpos.z);
}

// detekcia kolizie Avatara s objektami (trojuholnikmi) ********************
/*
 po stenach sa Avatar klze, s podlahou a stropom koliduje

 pouzite testy:

  1. test kolizie 3D bodu a trojuholnika
   poloha 3D bodu = poloha Avatara
   testovana vzdialenost = polovica sirky Avatara

  2. test kolizie 3D bodu so stranou trojuholnika
   p - v pripade kolizie novo vypocitana poloha Avatara
   q - v pripade kolizie 3D bod patriaci strane trojuholnika

  3. test kolizie usecky a trojuholnika
   ak Avatar nahle zmeni polohu o vacsiu vzdialenost ako je polovica
   jeho sirky, algoritmus kolizie 3D bodu a trojuholnika nebude fungovat
   spravne pretoze bod sa bude v jednom okamihu nachadzat pred trojuholnikom
   a v druhom za nim, pricom ziadna kolizia nebude detekovana, riesenim
   je doplnenie testu kolizie usecky a trojuholnika, ktory funguje vzdy spravne

 pouzite rovnice:

  rovnica roviny - AX+BY+CZ+D=0 => Y=-(AX+CZ+D)/B
  rovnica gule - (X-Cx)^2+(Y-Cy)^2+(Z-Cz)^2=R^2 => Y=Cy+sqrt(R^2-(X-Cx)^2-(Z-Cz)^2)
*/
void CAvatar::detectCollision(void)
{
 // nova (aktualna) a povodna poloha Avatara
 Apos=-actpos; Opos=-oldpos;

 // pocet trojuholnikov, pre ktore budu vykonane vypocty
 testNum=0;
 // indikatory detekovanych kolizii
 isWallCollision=false; isFloorCollision=false; isRoofCollision=false;

 // pre vsetky trojuholniky - 1. prechod - testuje sa podlaha
 loopvk(tris) {
  // testuje sa iba s trojuholnikmi predstavujucimi podlahu
  if (tris[k].n.y<avatarSlope) continue;
  // vzdialenost bodu od roviny definovanej trojuholnikom, pocita sa s novou polohou Avatara
  d1=Apos.x*tris[k].n.x+Apos.y*tris[k].n.y+Apos.z*tris[k].n.z+tris[k].d;
  // ak je bod pred trojuholnikom
  if (d1>=0.0f) {
   // ak je vzdialenost bodu od roviny vacsia ako testovana vzdialenost, nemohlo dojst ku kolizii
   if (d1>avatarWidth) continue;
   testNum++; // bude sa testovat - bod a trojuholnik
   // 1. test kolizie 3D bodu a trojuholnika
   // projekcia bodu do roviny definovanej trojuholnikom
   p=Apos-tris[k].n*d1;
   // test polohy bodu a trojuholnika
   if (m3d.point_in_triangle(p,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
    // bod lezi v trojuholniku
    // Y hodnota polohy Avatara sa vypocita dosadenim hodnot X a Z do rovnice roviny
    Apos.y=-(Apos.x*tris[k].n.x+Apos.z*tris[k].n.z+tris[k].d-avatarWidth)/tris[k].n.y;
    isFloorCollision=true; continue; // test pokracuje na dalsom trojuholniku
   }
   // 2. test kolizie 3D bodu so stranou trojuholnika
   p=Apos; q=Apos;
   if (m3d.edge_collision(&p,&q,tris[k].v[0],tris[k].v[1],tris[k].v[2],avatarWidth)) {
    // bod lezi na strane trojuholnika
    // Y hodnota polohy Avatara sa vypocita dosadenim hodnot X a Z do rovnice gule
    if (Opos.y>=Apos.y) { // iba ak Avatar pada
     Apos.y=q.y+sqrtf(avatarWidth*avatarWidth-(Apos.x-q.x)*(Apos.x-q.x)-(Apos.z-q.z)*(Apos.z-q.z));
     isFloorCollision=true; continue; // test pokracuje na dalsom trojuholniku
    }
   }
  } // if (d1>=0.0f)
  // ak je bod za trojuholnikom
  else {
   // 3. test kolizie usecky a trojuholnika
   // vzdialenost bodu od roviny definovanej trojuholnikom, pocita sa s povodnou polohou Avatara
   d2=Opos.x*tris[k].n.x+Opos.y*tris[k].n.y+Opos.z*tris[k].n.z+tris[k].d;
   // ku kolizii mohlo dojst iba ak sa Avatar pred zmenou polohy nachadzal
   // pred rovinou definovanou trojuholnikom
   if (d2<0.0f) continue;
   testNum++; // bude sa testovat - usecka a trojuholnik
   // p = vektor povodna poloha -> nova poloha Avatara
   p=Apos-Opos; p.normalize();
   // skalarny sucin vektora p a normaly trojuholnika
   d1=p.dot(tris[k].n);
   // projekcia bodu do roviny definovanej trojuholnikom
   // vzdialenost bodu od roviny / skalarny sucin vektorov
   d1=-d2/d1;
   // priesecnik vektora p a roviny definovanej trojuholnikom
   q=Opos+p*d1;
   // ak priesecnik lezi v trojuholniku, doslo ku kolizii
   if (m3d.point_in_triangle(q,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
    // priesecnik lezi v trojuholniku
    // Y hodnota polohy Avatara sa vypocita dosadenim hodnot X a Z do rovnice roviny
    Apos.y=-(Apos.x*tris[k].n.x+Apos.z*tris[k].n.z+tris[k].d-avatarWidth)/tris[k].n.y;
    isFloorCollision=true; continue; // test pokracuje na dalsom trojuholniku
   }
   // 2. test kolizie 3D bodu so stranou trojuholnika
   p=Apos; q=Apos;
   if (m3d.edge_collision(&p,&q,tris[k].v[0],tris[k].v[1],tris[k].v[2],avatarWidth)) {
    // bod lezi na strane trojuholnika
    // Y hodnota polohy Avatara sa vypocita dosadenim hodnot X a Z do rovnice gule
    if (Opos.y>=Apos.y) { // iba ak Avatar pada
     Apos.y=q.y+sqrtf(avatarWidth*avatarWidth-(Apos.x-q.x)*(Apos.x-q.x)-(Apos.z-q.z)*(Apos.z-q.z));
     isFloorCollision=true; continue; // test pokracuje na dalsom trojuholniku
    }
   }
  } // bod za trojuholnikom
 } // pre vsetky trojuholniky

 // pre vsetky trojuholniky - 2. prechod - testuju sa steny a strop
 loopvk(tris) {
  // testuje sa iba s trojuholnikmi predstavujucimi steny a strop
  if (tris[k].n.y>=avatarSlope) continue;

  // testuje sa bod na pozicii noh Avatara ---------------------------------

  // vzdialenost bodu od roviny definovanej trojuholnikom, pocita sa s novou polohou Avatara
  d1=Apos.x*tris[k].n.x+Apos.y*tris[k].n.y+Apos.z*tris[k].n.z+tris[k].d;
  // ak je bod pred trojuholnikom
  if (d1>=0.0f) {
   // ak je vzdialenost bodu od roviny mensia ako testovana vzdialenost, mohlo dojst ku kolizii
   if (d1<=avatarWidth) {
    testNum++; // bude sa testovat - bod a trojuholnik
    // 1. test kolizie 3D bodu a trojuholnika
    // projekcia bodu do roviny definovanej trojuholnikom
    p=Apos-tris[k].n*d1;
    // test polohy bodu a trojuholnika
    if (m3d.point_in_triangle(p,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
     // bod lezi v trojuholniku
     // Avatar sa posunie v smere normaly trojuholnika (efekt klzania po koliznej rovine)
     Apos=p+tris[k].n*avatarWidth;
     isWallCollision=true;
    }
    // 2. test kolizie 3D bodu so stranou trojuholnika
    p=Apos; q=Apos;
    if (m3d.edge_collision(&p,&q,tris[k].v[0],tris[k].v[1],tris[k].v[2],avatarWidth)) {
     // bod lezi na strane trojuholnika
     // Avatar sa posunie v smere "odstrediveho" vektora (efekt klzania po koliznej strane)
     Apos=p;
     isWallCollision=true;
    }
   } // if (d1<=avatarWidth)
  } // if (d1>=0.0f)
  // ak je bod za trojuholnikom
  else {
   // 3. test kolizie usecky a trojuholnika
   // vzdialenost bodu od roviny definovanej trojuholnikom, pocita sa s povodnou polohou Avatara
   d2=Opos.x*tris[k].n.x+Opos.y*tris[k].n.y+Opos.z*tris[k].n.z+tris[k].d;
   // ku kolizii mohlo dojst iba ak sa Avatar pred zmenou polohy nachadzal
   // pred rovinou definovanou trojuholnikom
   if (d2>=0.0f) {
    testNum++; // bude sa testovat - usecka a trojuholnik
    // p = vektor povodna poloha -> nova poloha Avatara
    p=Apos-Opos; p.normalize();
    // skalarny sucin vektora p a normaly trojuholnika
    d1=p.dot(tris[k].n);
    // projekcia bodu do roviny definovanej trojuholnikom
    // vzdialenost bodu od roviny / skalarny sucin vektorov
    d1=-d2/d1;
    // priesecnik vektora p a roviny definovanej trojuholnikom
    q=Opos+p*d1;
    // ak priesecnik lezi v trojuholniku, doslo ku kolizii
    if (m3d.point_in_triangle(q,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
     // priesecnik lezi v trojuholniku
     // Avatar sa posunie v smere normaly trojuholnika (efekt klzania po koliznej rovine)
     Apos=q+tris[k].n*avatarWidth;
     isWallCollision=true;
    }
    // 2. test kolizie 3D bodu so stranou trojuholnika
    p=Apos; q=Apos;
    if (m3d.edge_collision(&p,&q,tris[k].v[0],tris[k].v[1],tris[k].v[2],avatarWidth)) {
     // bod lezi na strane trojuholnika
     // Avatar sa posunie v smere "odstrediveho" vektora (efekt klzania po koliznej strane)
     Apos=p;
     isWallCollision=true;
    }
   } // if (d2>=0.0f)
  } // bod za trojuholnikom

  // testuje sa bod na pozicii oci Avatara ---------------------------------

  ap=Apos; ap.y+=crouch*avatarEye; op=Opos; op.y+=crouch*avatarEye;

  // vzdialenost bodu od roviny definovanej trojuholnikom, pocita sa s novou polohou Avatara
  d1=ap.x*tris[k].n.x+ap.y*tris[k].n.y+ap.z*tris[k].n.z+tris[k].d;
  // ak je bod pred trojuholnikom
  if (d1>=0.0f) {
   // ak je vzdialenost bodu od roviny vacsia ako testovana vzdialenost, nemohlo dojst ku kolizii
   if (d1>avatarWidth) continue;
   testNum++; // bude sa testovat - bod a trojuholnik
   // 1. test kolizie 3D bodu a trojuholnika
   // projekcia bodu do roviny definovanej trojuholnikom
   p=ap-tris[k].n*d1;
   // test polohy bodu a trojuholnika
   if (m3d.point_in_triangle(p,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
    // bod lezi v trojuholniku
    // Avatar sa posunie v smere normaly trojuholnika (efekt klzania po koliznej rovine)
    Apos=p+tris[k].n*avatarWidth; Apos.y-=crouch*avatarEye;
    if (tris[k].n.y>-avatarSlope) isWallCollision=true; else isRoofCollision=true;
    continue; // test pokracuje na dalsom trojuholniku
   }
   // 2. test kolizie 3D bodu so stranou trojuholnika
   p=ap; q=ap;
   if (m3d.edge_collision(&p,&q,tris[k].v[0],tris[k].v[1],tris[k].v[2],avatarWidth)) {
    // bod lezi na strane trojuholnika
    // Avatar sa posunie v smere "odstrediveho" vektora (efekt klzania po koliznej strane)
    Apos=p; Apos.y-=crouch*avatarEye;
    if (tris[k].n.y>-avatarSlope) isWallCollision=true; else isRoofCollision=true;
    continue; // test pokracuje na dalsom trojuholniku
   }
  } // if (d1>=0.0f)
  // ak je bod za trojuholnikom
  else {
   // 3. test kolizie usecky a trojuholnika
   // vzdialenost bodu od roviny definovanej trojuholnikom, pocita sa s povodnou polohou Avatara
   d2=op.x*tris[k].n.x+op.y*tris[k].n.y+op.z*tris[k].n.z+tris[k].d;
   // ku kolizii mohlo dojst iba ak sa Avatar pred zmenou polohy nachadzal
   // pred rovinou definovanou trojuholnikom
   if (d2<0.0f) continue;
   testNum++; // bude sa testovat - usecka a trojuholnik
   // p = vektor povodna poloha -> nova poloha Avatara
   p=ap-op; p.normalize();
   // skalarny sucin vektora p a normaly trojuholnika
   d1=p.dot(tris[k].n);
   // projekcia bodu do roviny definovanej trojuholnikom
   // vzdialenost bodu od roviny / skalarny sucin vektorov
   d1=-d2/d1;
   // priesecnik vektora p a roviny definovanej trojuholnikom
   q=op+p*d1;
   // ak priesecnik lezi v trojuholniku, doslo ku kolizii
   if (m3d.point_in_triangle(q,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
    // priesecnik lezi v trojuholniku
    // Avatar sa posunie v smere normaly trojuholnika (efekt klzania po koliznej rovine)
    Apos=q+tris[k].n*avatarWidth; Apos.y-=crouch*avatarEye;
    if (tris[k].n.y>-avatarSlope) isWallCollision=true; else isRoofCollision=true;
    continue; // test pokracuje na dalsom trojuholniku
   }
   // 2. test kolizie 3D bodu so stranou trojuholnika
   p=ap; q=ap;
   if (m3d.edge_collision(&p,&q,tris[k].v[0],tris[k].v[1],tris[k].v[2],avatarWidth)) {
    // bod lezi na strane trojuholnika
    // Avatar sa posunie v smere "odstrediveho" vektora (efekt klzania po koliznej strane)
    Apos=p; Apos.y-=crouch*avatarEye;
    if (tris[k].n.y>-avatarSlope) isWallCollision=true; else isRoofCollision=true;
    continue; // test pokracuje na dalsom trojuholniku
   }
  } // bod za trojuholnikom
 } // pre vsetky trojuholniky

 // INFO: pocet trojuholnikov, s ktorymi sa testovala kolizia
 //printf("%u\n",testNum);

 // v pripade kolizie sa Avatar umiestni na novu polohu,
 // ak kolizia nenastala, poloha Avatara vypocitana v update() sa nezmeni
 actpos=-Apos;
}

// detekcia kolizie Avatara so stropom v okamihu ked Avatar cupi a chce vstat
void CAvatar::detectRoofCollision(void)
{
 // poloha noh a poloha oci Avatara v okamihu, ked Avatar vstane
 op=-actpos; ap=op; ap.y+=crouch*avatarEye;

 isRoofCollision=false;

 // pre vsetky trojuholniky
 loopvk(tris) {
  // testuje sa iba s trojuholnikmi predstavujucimi strop
  if (tris[k].n.y>-avatarSlope) continue;
  // test kolizie usecky a trojuholnika
  d1=op.x*tris[k].n.x+op.y*tris[k].n.y+op.z*tris[k].n.z+tris[k].d;
  d2=ap.x*tris[k].n.x+ap.y*tris[k].n.y+ap.z*tris[k].n.z+tris[k].d;
  if (d1*d2>0.0f) continue;
  // p = vektor poloha noh -> poloha oci
  p=ap-op; p.normalize();
  // skalarny sucin vektora p a normaly trojuholnika
  d1=p.dot(tris[k].n);
  // projekcia bodu do roviny definovanej trojuholnikom
  // vzdialenost bodu od roviny / skalarny sucin vektorov
  d1=-d2/d1;
  // priesecnik vektora p a roviny definovanej trojuholnikom
  q=op+p*d1;
  // ak priesecnik lezi v trojuholniku, doslo ku kolizii
  if (m3d.point_in_triangle(q,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
   isRoofCollision=true; break;
  }
 }
}
