#ifndef _AVATAR_H_
#define _AVATAR_H_

// trieda premennych a funkcii (objektov) pre pracu s Avatarom
class CAvatar {
public:
 CAvatar();  // konstruktor
 ~CAvatar(); // destruktor

 // vyska oci Avatara
 GLfloat avatarEye;
 // polovica sirky Avatara
 GLfloat avatarWidth;
 // sklon roviny, ktory dokaze Avatar prekonat
 GLfloat avatarSlope;

 vec3 actpos; // aktualna poloha Avatara
 vec3 eyepos; // aktualna poloha oci Avatara

 vec3 targetV; // vektor smerujuci dopredu
 vec3 rightV;  // vektor smerujuci doprava

 // indikatory pre vykonavanie definovanych posunov Avatara
 GLfloat goForward,goBack,goUp,goDown,goRight,goLeft;
 GLfloat goLeanLeft,goLeanRight;
 GLfloat goCrouch;
 GLfloat goJump,goFast;
 GLubyte isFire1; // strelba1 (lave tlacidlo mysi)
 GLubyte isFire2; // strelba2 (prave tlacidlo mysi)

 // inicializacia Avatara
 void init(void);
 // nastavenie natocenia
 void set_rotate(float a,float b,float c) { rotate.set(a,b,c); }
 // reset parametrov Avatara
 void reset(void);
 // nastavenie prirastkov pre zmenu polohy Avatara
 void setmovxyz(void);
 // nastavenie prirastkov pre zmenu natocenia Avatara
 void setrotxy(GLfloat X,GLfloat Y);
 void setrotzl(GLfloat Z);
 void setrotzr(GLfloat Z);
 // aktualizacia parametrov Avatara
 void update(void);

 // Avatar sa hybe
 bool is_moving(void) { return((accelX!=0.0f)||(accelZ!=0.0f)); }
 // Avatar cupi
 bool is_crouching(void) { return(crouch!=1.0f); }
 // Avatar skace
 bool is_jumping(void) { if (jumpTime!=0.0f) return(true); else return(false); }
 // natocenie Avatara
 vec3 get_angle(void) { return(angle); }
private:
 GLubyte movetype; // typ pohybu

 mat16 mat; // transformacna matica Avatara, hodnoty prvkov hlavnej diagonaly
            // sa automaticky nastavia na 1.0f (v math3d.h)

 vec3 oldpos,move,oldmove; // povodna poloha Avatara, prirastok pre zmenu polohy
 vec3 angle,rotate;        // natocenie, prirastok pre zmenu natocenia

 GLfloat tilt,tiltAbs; // kracanie Avatara
 bool goTilt;
 GLfloat lean,leanAdd; // uklon Avatara od zvislej roviny

 GLfloat canRotate;    // indikator dokoncenia rotacie

 // koeficienty
 GLfloat speed,slide,altit; // rychlost, bocny pohyb, zmena vysky
 GLfloat gravity;           // gravitacne zrychlenie
 GLfloat viewnear,viewfar;  // minimalny a maximalny dohlad
 GLfloat fov;               // zorny uhol pohladu
 GLuint jumpSpeed;          // rychlost skoku
 GLfloat jumpTime;          // cas zacatia skoku

 GLfloat accelX,accelZ; // akceleracia Avatara
 GLfloat accelAddX,accelAddZ;

 GLfloat rightLeft,forwardBack; // tlacidla vlavo/vpravo, dopredu/dozadu

 GLfloat gravityTime;     // aktualny cas posobenia gravitacie
 GLfloat gT;              // prirastok casu pri skoku/posobeni gravitacie - zavisly na FPS
 GLfloat gKoef[MAXFPS+1]; // prirastky casu pre skok/pad - zavisle na FPS

 GLfloat crouch,crouchAdd; // Avatar stoji/cupi

 GLfloat d1,d2,d3; // pomocne premenne
 vec3 p,q;         // pomocne vektory
 vec3 ap,op;       // pomocne premenne

 vec3 Apos,Opos; // pomocne vektory pre docasne ulozenie poloh Avatara

 GLuint testNum; // pocet trojuholnikov, s ktorymi sa testuje kolizia
 // indikatory detekovanych kolizii
 bool isWallCollision,isFloorCollision,isRoofCollision;

 // detekcia kolizie Avatara s objektami (trojuholnikmi)
 void detectCollision(void);
 // detekcia kolizie Avatara so stropom v okamihu ked Avatar cupi a chce vstat
 void detectRoofCollision(void);
};

#endif
