#ifndef _MATH3D_H_
#define _MATH3D_H_

#include <cstring>
#include <cstdlib>
#include <cmath>

// rozne geometricke a matematicke funkcie *********************************

// minimum a maximum 2 hodnot
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif
template <class T>
static inline T min(const T a,const T b) { return(a < b ? a : b); }
template <class T>
static inline T min3(const T a,const T b,const T c) { return(min(a,min(b,c))); }
template <class T>
static inline T min4(const T a,const T b,const T c,const T d) { return (min(a,min3(b,c,d))); }
template <class T>
static inline T max(const T a,const T b) { return(a > b ? a : b); }
template <class T>
static inline T max3(const T a,const T b,const T c) { return(max(a,max(b,c))); }
template <class T>
static inline T max4(const T a,const T b,const T c,const T d) { return (max(a,max3(b,c,d))); }

#ifdef swap
#undef swap
#endif
// vymena hodnot 2 premennych
template <class T>
static inline void swap(T &a,T &b) { T t=a; a=b; b=t; }

#ifdef ranged_random
#undef ranged_random
#endif
inline float ranged_random(float A,float B) { return(A+(B-A)*((float) rand())/((float) RAND_MAX)); }

#ifdef square
#undef square
#endif
template<typename T>
T square(T a) { return(a*a); }

// konstanty
#ifndef PI
#define PI 3.1415926535897932384626433832795
#endif
#define TWO_PI 6.28318530717958647692528676655901
#define PI180 (0.017453292519f)
#define PI_DIV_180 0.0174532925199432957692369076848861
#define _180_DIV_PI 57.2957795130823208767981548141052
#define PI2   (1.570796326794f)
// prevod stupne<->radiany
#define DEG2RAD(a) (PI_DIV_180*(a))
#define RAD2DEG(a) (_180_DIV_PI*(a))

inline float sin_deg(float deg) { return (float) sin(DEG2RAD(deg)); }
inline float cos_deg(float deg) { return (float) cos(DEG2RAD(deg)); }
inline float asin_deg(float x) { return RAD2DEG(asin((float) x)); }
inline float acos_deg(float x) { return RAD2DEG(acos((float) x)); }
inline float tan_deg(float deg) { return (float) tan(DEG2RAD(deg)); }
inline float atan2_deg(float x,float y) { return RAD2DEG((float) atan2(x,y)); }
inline float atan_deg(float x) { return RAD2DEG((float) atan(x)); }

// cislo v intervale (-100.000.000;100.000.000)
inline bool is_finite(float V) { return((V>-1e8)&&(V<1e8)); }

// absolutna hodnota
inline float my_abs(float V) { return(V > 0.0f ? V : -V); }

// inverzna - prevratena hodnota cisla
inline float inv_num(float N) { if (N>0.0f) return (1.0f/N); else return (1.0e30); }

// 2-rozmerny vektor *******************************************************
class vec2 {
 public:

  // udaje -----------------------------------------------------------------
  union {
   struct {
    float x,y;
   };
   float v[2];
  };

  // konstruktory ----------------------------------------------------------
  vec2(float A=0.0f,float B=0.0f):x(A),y(B) {}
  vec2(const vec2 &V):x(V.x),y(V.y) {}
  vec2(float *V):x(V[0]),y(V[1]) {}
  vec2(int V[2]):x(V[0]),y(V[1]) {}

  // operatory -------------------------------------------------------------
  // priradenie hodnot prvkom vektora
  void set(float A,float B) { x=A; y=B; }
  // scitanie vektorov
  const vec2 operator+(const vec2 &V) const { return(vec2(x+V.x,y+V.y)); }
  // odcitanie vektorov
  const vec2 operator-(const vec2 &V) const { return(vec2(x-V.x,y-V.y)); }
  // nasobenie skalarom
  const vec2 operator*(const float S) const { return(vec2(x*S,y*S)); }

  const float *get_data() { return(v); }
};

// 3-rozmerny vektor *******************************************************
class vec3 {
 public:

  // udaje -----------------------------------------------------------------
  union {
   struct {
    float x,y,z;
   };
   float v[3];
  };

  // konstruktory ----------------------------------------------------------
  //vec3(float A=0.0f,float B=0.0f,float C=0.0f):x(A),y(B),z(C) {}
  vec3() { x=0.0f; y=0.0f; z=0.0f; }
  vec3(float A,float B,float C) { x=A; y=B; z=C; }
  vec3(const vec3 &V):x(V.x),y(V.y),z(V.z) {}
  vec3(float *V):x(V[0]),y(V[1]),z(V[2]) {}
  vec3(int V[3]):x(V[0]),y(V[1]),z(V[2]) {}
  vec3(const vec3 &A,const vec3 &B):x(B.x-A.x),y(B.y-A.y),z(B.z-A.z) {}
  explicit vec3(int A):x(A),y(A),z(A) {}
  explicit vec3(float A):x(A),y(A),z(A) {}

  // operatory -------------------------------------------------------------
  // index
  float &operator[](int I)       { return v[I]; }
  float  operator[](int I) const { return v[I]; }
  // priradenie hodnot prvkom vektora
  const vec3 &operator=(const vec3 &V) { x=V.x; y=V.y; z=V.z; return(*this); }
  void set(float A,float B,float C) { x=A; y=B; z=C; }
  void set_to(float V) { v[0]=V; v[1]=V; v[2]=V; }
  vec3 &set(int I,float F) { v[I]=F; return(*this); }
  // rovnost vektorov
  const bool operator==(const vec3 &V) const { return((x==V.x) && (y==V.y) && (z==V.z)); }
  // nerovnost vektorov
  const bool operator!=(const vec3 &V) const { return(!(*this==V)); }
  // scitanie vektorov
  const vec3 operator+(const vec3 &V) const { return(vec3(x+V.x,y+V.y,z+V.z)); }
  vec3 &add(const vec3 &V) { x+=V.x; y+=V.y; z+=V.z; return(*this); }
  // pricitanie skalara k prvkom vektora
  vec3 &add(float S) { x+=S; y+=S; z+=S; return(*this); }
  // pricitanie k vektoru
  const vec3 &operator+=(const vec3 &V) { x+=V.x; y+=V.y; z+=V.z; return(*this); }
  // odcitanie vektorov
  const vec3 operator-(const vec3 &V) const { return(vec3(x-V.x,y-V.y,z-V.z)); }
  vec3 &sub(const vec3 &V) { x-=V.x; y-=V.y; z-=V.z; return(*this); }
  // odcitanie skalara od prvkov vektora
  vec3 &sub(float S) { x-=S; y-=S; z-=S; return(*this); }
  // odcitanie od vektora
  const vec3 &operator-=(const vec3 &V) { x-=V.x; y-=V.y; z-=V.z; return(*this); }
  // nasobenie skalarom
  const vec3 &operator*=(const float S) { x*=S; y*=S; z*=S; return(*this); }
  void component_product_update(const vec3 &V) { x*=V.x; y*=V.y; z*=V.z; }
  //const vec3 operator*(const float S) const { return(vec3(x*S,y*S,z*S)); }
  friend vec3 operator*(const vec3 &A,const float B);
  friend vec3 operator*(const float A,const vec3 &B);
  vec3 &mul(float F) { x*=F; y*=F; z*=F; return(*this); }
  // delenie skalarom
  const vec3 &operator/=(const float S) { const float recip=1.0f/S; x*=recip; y*=recip; z*=recip; return(*this); }
  //const vec3 operator/(const float S) const { const float recip=1.0f/S; return(vec3(x*recip,y*recip,z*recip)); }
  friend vec3 operator/(const vec3 &A,const float B);
  vec3 &div(float F) { if (F==0.0f) F=1.0f; x/=F; y/=F; z/=F; return(*this); }
  // nasobenie vektorov
  const vec3 operator*(const vec3 &V) const { return(vec3(x*V.x,y*V.y,z*V.z)); }
  vec3 component_product(const vec3 &V) const { return(vec3(x*V.x,y*V.y,z*V.z)); }
  // kladne znamienko vektora
  const vec3 operator+() const { return(vec3(*this)); }
  // zaporne znamienko vektora
  const vec3 operator-() const { return(vec3(-x,-y,-z)); }
  vec3 &neg() { return(mul(-1.0f)); }

  // funkcie ---------------------------------------------------------------
  // vynulovanie prvkov vektora
  void clear(void) { x=y=z=0.0f; }
  void invert() { x=-x; y=-y; z=-z; }
  bool sensible() const {
   for (unsigned int i=3;i--!=0;) { if (!is_finite(v[i])) return(false); }
   return(true);
  }
  const float *get_data() { return(v); }
  const float *get_data() const { return(v); }
  // vektorovy sucin
  const vec3 cross(const vec3 &V) const { return(vec3(y*V.z-z*V.y,z*V.x-x*V.z,x*V.y-y*V.x)); }
  vec3 &cross(const vec3 &V1,const vec3 &V2) {
   x=V1.y*V2.z-V1.z*V2.y; y=V1.z*V2.x-V1.x*V2.z; z=V1.x*V2.y-V1.y*V2.x;
   return(*this);
  }
  friend vec3 cross(const vec3 &A,const vec3 &B);
  // skalarny sucin
  float dot(const vec3 &V) const { return(x*V.x+y*V.y+z*V.z); }
  friend float dot(const vec3 &A,const vec3 &B);
  // (dlzka vektora)^2
  float squaredlen() const { return(x*x+y*y+z*z); }
  // dlzka vektora
  float magnitude() const { return(sqrtf(squaredlen())); }
  // jednotkovy vektor
  const vec3 unit() const { return((*this)/magnitude()); }
  // normalizacia vektora
  vec3 &normalize() { div(magnitude()); return(*this); }
  // uhol medzi vektormi
  const float angle(const vec3 &V) const { return(acosf(dot(V))); }
  // zrkadlova projekcia 3D bodu do roviny definovanej normalou
  // rovina prechadza zaciatkom suradnicovej sustavy
  const vec3 reflect(const vec3 &N) const { return(*this-N*2.0f*dot(N)); }
  vec3 &reflect(const vec3 &N) { float k=2.0f*dot(N); x-=N.x*k; y-=N.y*k; z-=N.z*k; return(*this); }
  // rotacia vektora okolo normaly
  const vec3 rotate(const float A,const vec3 &N) const {
   const float cosine=cosf(A); const float sine=sinf(A);
   return(vec3((*this)*cosine+((N*(*this))*(1.0f-cosine))*N+(cross(N))*sine));
  }
  // nulovy vektor?
  bool iszero() const { return(x==0 && y==0 && z==0); }
  // normalizovany vektor?
  bool isnormalized() const { float m=squaredlen(); return(m>0.99f && m<1.01f); }
  // vzdialenost dvoch 3D bodov
  float dist(const vec3 &V) const { vec3 t; return(dist(V,t)); }
  float dist(const vec3 &V1,vec3 &V2) const { V2=(*this); V2.sub(V1); return(V2.magnitude()); }
  // reject
  bool reject(const vec3 &V,float MAX) { return(x>V.x+MAX || x<V.x-MAX || y>V.y+MAX || y<V.y-MAX); }
  // rescale
  void rescale(float K) { float mag=magnitude(); if (mag>1e-6f) mul(K/mag); }
  // projekcia 3D bodu do roviny
  // rovina prechadza zaciatkom suradnicovej sustavy
  vec3 &project(const vec3 &N) { float k=dot(N); x-=k*N.x; y-=k*N.y; z-=k*N.z; return(*this); }
  vec3 &projectxydir(const vec3 &N) { if (N.z) z=-(x*N.x/N.z+y*N.y/N.z); return(*this); }
  vec3 &projectxy(const vec3 &N) {
   float m=squaredlen(),k=dot(N);
   projectxydir(N);
   rescale(sqrtf(max(m-k*k,0.0f)));
   return(*this);
  }
  vec3 &projectxy(const vec3 &N,float THRESHOLD) {
   float m=squaredlen(),k=min(dot(N),THRESHOLD);
   projectxydir(N);
   rescale(sqrtf(max(m-k*k,0.0f)));
   return(*this);
  }
  vec3 proj(const vec3 &A,const vec3 &B) {
   float a;
   a=A.dot(B);
   return((A*a)/A.squaredlen());
  }
  // rotacia okolo osi X Y Z
  void rotate_around_x(float A) { *this=vec3(x,                  cosf(A)*y-sinf(A)*z,cosf(A)*z+sinf(A)*y); }
  void rotate_around_y(float A) { *this=vec3(cosf(A)*x-sinf(A)*z,y,                  cosf(A)*z+sinf(A)*x); }
  void rotate_around_z(float A) { *this=vec3(cosf(A)*x-sinf(A)*y,cosf(A)*y+sinf(A)*x,z); }
  // rotacia okolo osi O
  void rotate(float A,const vec3 &O) {
   float c=cosf(A),s=sinf(A);
   rotate(c,s,O);
  }
  void rotate(float C,float S,const vec3 &O) {
   *this=vec3(x*(O.x*O.x*(1-C)+C)    +y*(O.x*O.y*(1-C)-O.z*S)+z*(O.x*O.z*(1-C)+O.y*S),
              x*(O.y*O.x*(1-C)+O.z*S)+y*(O.y*O.y*(1-C)+C)    +z*(O.y*O.z*(1-C)-O.x*S),
              x*(O.x*O.z*(1-C)-O.y*S)+y*(O.y*O.z*(1-C)+O.x*S)+z*(O.z*O.z*(1-C)+C));
  }
  void show(const char *S) const {
   for (unsigned int i=0;i<3;i++) {}
  }
};
// globalne operatory a funkcie suvisiace s triedou vec3 *******************
inline vec3 operator*(const vec3 &A,const float B) { return(vec3(A.x*B,A.y*B,A.z*B)); }
inline vec3 operator*(const float A,const vec3 &B) { return(B*A); }
inline vec3 operator/(const vec3 &A,const float B) {
 const float recip=1.0f/B; return(vec3(A.x*recip,A.y*recip,A.z*recip));
}
inline vec3 cross(const vec3 &A,const vec3 &B) {
 return(vec3(A.y*B.z-A.z*B.y,A.z*B.x-A.x*B.z,A.x*B.y-A.y*B.x));
}
inline float dot(const vec3 &A,const vec3 &B) {
 return(A.x*B.x+A.y*B.y+A.z*B.z);
}

// 4-rozmerny vektor *******************************************************
class vec4 {
 public:

  // udaje -----------------------------------------------------------------
  union {
   struct {
    float x,y,z,w;
   };
   float v[4];
  };

  // konstruktory ----------------------------------------------------------
  vec4(float A=0.0f,float B=0.0f,float C=0.0f,float D=0.0f):x(A),y(B),z(C),w(D) {}
  vec4(const vec4 &V):x(V.x),y(V.y),z(V.z),w(V.w) {}
  vec4(float *V):x(V[0]),y(V[1]),z(V[2]),w(V[3]) {}
  vec4(int V[4]):x(V[0]),y(V[1]),z(V[2]),w(V[3]) {}

  // operatory -------------------------------------------------------------
  // index
  float &operator[](int I)       { return v[I]; }
  float  operator[](int I) const { return v[I]; }
  // priradenie hodnot prvkom vektora
  const vec4 &operator=(const vec4 &V) { x=V.x; y=V.y; z=V.z; w=V.w; return(*this); }
  void set(float A,float B,float C,float D) { x=A; y=B; z=C; w=D; }

  const float *get_data() const { return(v); }
};

// quaternion **************************************************************
class CQuaternion {
 public:
  union {
   struct {
    float r,i,j,k;
   };
   float d[4];
  };
  // konstruktory
  CQuaternion():r(1),i(0),j(0),k(0) {}
  CQuaternion(const float R,const float I,const float J,const float K)
   :r(R),i(I),j(J),k(K) {}
  // normalizacia = quaternion bude platny
  void normalize(void) {
   float d=(r*r+i*i+j*j+k*k);
   if (d==0.0f) { r=1.0f; return; }
   d=1.0f/sqrt(d);
   r*=d; i*=d; j*=d; k*=d;
  }
  // nasobenie quaternionov
  void operator*=(const CQuaternion &B) {
   CQuaternion a=*this;
   r=a.r*B.r-a.i*B.i-a.j*B.j-a.k*B.k;
   i=a.r*B.i+a.i*B.r+a.j*B.k-a.k*B.j;
   j=a.r*B.j+a.j*B.r+a.k*B.i-a.i*B.k;
   k=a.r*B.k+a.k*B.r+a.i*B.j-a.j*B.i;
  }
  // pricitanie vektora ku quaterniounu
  void add_vector(const vec3 V) {
   CQuaternion a(0.0f,V.x,V.y,V.z);
   a*=(*this);
   r+=a.r*(0.5f);
   i+=a.i*(0.5f);
   j+=a.j*(0.5f);
   k+=a.k*(0.5f);
  }
  // pricitanie skalovaneho vektora ku quaterniounu
  void add_scaled_vector(const vec3 V,float S) {
   CQuaternion a(0,V.x*S,V.y*S,V.z*S);
   a*=*this;
   r+=a.r*(0.5f);
   i+=a.i*(0.5f);
   j+=a.j*(0.5f);
   k+=a.k*(0.5f);
  }
  void rotate_by_vector(const vec3 &V) {
   CQuaternion a(0,V.x,V.y,V.z);
   (*this)*=a;
  }
};

// matica typu 3x3 *********************************************************
class mat3x3 {
public:
 float d[9];

 // konstruktory -----------------------------------------------------------
 mat3x3() { d[0]=d[1]=d[2]=d[3]=d[4]=d[5]=d[6]=d[7]=d[8]=0.0f; }
 mat3x3(float v11,float v21,float v31,
        float v12,float v22,float v32,
        float v13,float v23,float v33) {
  d[0]=v11; d[1]=v21; d[2]=v31;
  d[3]=v12; d[4]=v22; d[5]=v32;
  d[6]=v13; d[7]=v23; d[8]=v33;
 }
 mat3x3(const vec3 &v1,const vec3 &v2,const vec3 &v3) {
  d[0]=v1[0]; d[1]=v1[1]; d[2]=v1[2];
  d[3]=v2[0]; d[4]=v2[1]; d[5]=v2[2];
  d[6]=v3[0]; d[7]=v3[1]; d[8]=v3[2];
 }
 void set_to(float V) {
  d[0]=V; d[1]=V; d[2]=V;
  d[3]=V; d[4]=V; d[5]=V;
  d[6]=V; d[7]=V; d[8]=V;
 }
 void set_diagonal(float A,float B,float C) { set_inertia_tensor(A,B,C); }
 explicit inline mat3x3(float V) { set_to(V); }
 // destruktor -------------------------------------------------------------
 ~mat3x3() {}

 // operatory --------------------------------------------------------------
 float &operator()(const unsigned int I,const unsigned int J) { return(d[I+J*3]); }
 const float &operator()(const unsigned int I,const unsigned int J) const { return(d[I+J*3]); }

 mat3x3 &operator+=(const mat3x3 &D) {
  loopi(9) d[i]+=D.d[i];
  return(*this);
 }
 mat3x3 &operator-=(const mat3x3 &D) {
  loopi(9) d[i]-=D.d[i];
  return(*this);
 }
 mat3x3 &operator*=(const float D) {
  loopi(9) d[i]*=D;
  return(*this);
 }
 mat3x3 &operator/=(const float D) {
  const float inv_D=1.0f/D;
  loopi(9) d[i]*=inv_D;
  return(*this);
 }
 mat3x3 operator+(const mat3x3 &D) const { return(mat3x3(*this)+=D); }
 mat3x3 operator-(const mat3x3 &D) const { return(mat3x3(*this)-=D); }
 friend mat3x3 operator*(const mat3x3 &A,const float B);
 friend mat3x3 operator*(const float A,const mat3x3 &B);
 friend mat3x3 operator*(const mat3x3 &A,const mat3x3 &B);
 vec3 operator*(const vec3 &V) const {
  return(vec3(V.x*d[0]+V.y*d[1]+V.z*d[2],
              V.x*d[3]+V.y*d[4]+V.z*d[5],
              V.x*d[6]+V.y*d[7]+V.z*d[8]));
 }
 void operator*=(const mat3x3 &M) {
  float t1,t2,t3;

  t1=d[0]*M.d[0]+d[1]*M.d[3]+d[2]*M.d[6];
  t2=d[0]*M.d[1]+d[1]*M.d[4]+d[2]*M.d[7];
  t3=d[0]*M.d[2]+d[1]*M.d[5]+d[2]*M.d[8];
  d[0]=t1; d[1]=t2; d[2]=t3;

  t1=d[3]*M.d[0]+d[4]*M.d[3]+d[5]*M.d[6];
  t2=d[3]*M.d[1]+d[4]*M.d[4]+d[5]*M.d[7];
  t3=d[3]*M.d[2]+d[4]*M.d[5]+d[5]*M.d[8];
  d[3]=t1; d[4]=t2; d[5]=t3;

  t1=d[6]*M.d[0]+d[7]*M.d[3]+d[8]*M.d[6];
  t2=d[6]*M.d[1]+d[7]*M.d[4]+d[8]*M.d[7];
  t3=d[6]*M.d[2]+d[7]*M.d[5]+d[8]*M.d[8];
  d[6]=t1; d[7]=t2; d[8]=t3;
 }

 // funkcie ----------------------------------------------------------------
 bool sensible() const {
  loopi(9) {
   if (!is_finite(d[i])) return(false);
  }
  return(true);
 }
 void set_inertia_tensor(float Ix,float Iy,float Iz,
                         float Ixy=0.0f,float Ixz=0.0f,float Iyz=0.0f) {
  d[0]=Ix;
  d[1]=d[3]=-Ixy;
  d[2]=d[6]=-Ixz;
  d[4]=Iy;
  d[5]=d[7]=-Iyz;
  d[8]=Iz;
 }
 void set_box_inertia_tensor(const vec3 &S,float M) {
  vec3 squares=S.component_product(S);
  set_inertia_tensor((1.0f/3.0f)*M*(squares.y+squares.z),
                     (1.0f/3.0f)*M*(squares.x+squares.z),
                     (1.0f/3.0f)*M*(squares.x+squares.y));
 }
 void set_skew_symetric(const vec3 V) {
  d[0]=d[4]=d[8]=0.0f;
  d[1]=-V.z; d[2]= V.y;
  d[3]= V.z; d[5]=-V.x;
  d[6]=-V.y; d[7]= V.x;
 }
 void set_components(const vec3 &A,const vec3 &B,const vec3 &C) {
  d[0]=A.x; d[1]=B.x; d[2]=C.x;
  d[3]=A.y; d[4]=B.y; d[5]=C.y;
  d[6]=A.z; d[7]=B.z; d[8]=C.z;
 }
 vec3 mul(const vec3 &V) const { return((*this)*V); }
 vec3 mul_T(const vec3 &V) const {
  return(vec3(V.x*d[0]+V.y*d[3]+V.z*d[6],
              V.x*d[1]+V.y*d[4]+V.z*d[7],
              V.x*d[2]+V.y*d[5]+V.z*d[8]));
 }
 vec3 get_row_vector(int I) const { return(vec3(d[I*3],d[I*3+1],d[I*3+2])); }
 vec3 get_column_vector(int I) const { return(vec3(d[I],d[I+3],d[I+6])); }
 void set_inverse(const mat3x3 M) {
  // determinant matice 2x2
  // a11 a12  0  1
  // a21 a22  2  3
  // detM=a11 a22 - a12 a21 = 0 3 - 1 2
  //
  // Sarusovo (krizove) pravidlo pre vypocet determinantu matice 3x3
  // a11 a12 a13  0  1  2
  // a21 a22 a23  3  4  5
  // a31 a32 a33  6  7  8
  // detM = a11 a22 a33 + a21 a32 a13 + a31 a12 a23 -
  //        a13 a22 a31 - a23 a32 a11 - a33 a12 a21
  //      = 0 4 8 + 3 7 2 + 6 1 5 -
  //        2 4 6 - 5 7 0 - 8 1 3
  float detM=(M.d[0]*M.d[4]*M.d[8]+M.d[3]*M.d[7]*M.d[2]+M.d[6]*M.d[1]*M.d[5]-
              M.d[2]*M.d[4]*M.d[6]-M.d[5]*M.d[7]*M.d[0]-M.d[8]*M.d[1]*M.d[3]);
  if (detM==0.0f) return;
  detM=1.0f/detM;

  // Mij - matica, ktora vznikla z povodnej matice vynechanim
  // i-teha riadku a j-teho stlpca
  // Aij = (-1)^(i+j) * detMij - algebraicky doplnok matice
  // A_adj - adjungovana matica zlozena z algebraickych doplnkov
  // algebraicke doplnky su v adjungovanej matici ulozene transponovane
  // prvok ij adjungovanej matice = Aji
  //
  // M_inv = A_adj / detM - iverzna matica
  d[0]= (M.d[4]*M.d[8]-M.d[5]*M.d[7])*detM;
  d[1]=-(M.d[1]*M.d[8]-M.d[2]*M.d[7])*detM;
  d[2]= (M.d[1]*M.d[5]-M.d[2]*M.d[4])*detM;
  d[3]=-(M.d[3]*M.d[8]-M.d[5]*M.d[6])*detM;
  d[4]= (M.d[0]*M.d[8]-M.d[2]*M.d[6])*detM;
  d[5]=-(M.d[0]*M.d[5]-M.d[2]*M.d[3])*detM;
  d[6]= (M.d[3]*M.d[7]-M.d[4]*M.d[6])*detM;
  d[7]=-(M.d[0]*M.d[7]-M.d[1]*M.d[6])*detM;
  d[8]= (M.d[0]*M.d[4]-M.d[1]*M.d[3])*detM;
 }
 mat3x3 inverse() const {
  mat3x3 result;
  result.set_inverse(*this);
  return(result);
 }
 void invert() { set_inverse(*this); }
 void set_transpose(const mat3x3 &M) {
  d[0]=M.d[0]; d[1]=M.d[3]; d[2]=M.d[6];
  d[3]=M.d[1]; d[4]=M.d[4]; d[5]=M.d[7];
  d[6]=M.d[2]; d[7]=M.d[5]; d[8]=M.d[8];
 }
 mat3x3 transpose() const {
  mat3x3 result;
  result.set_transpose(*this);
  return(result);
 }
 void set_orientation(const CQuaternion &Q) {
  d[0]=1.0f-(2.0f*Q.j*Q.j+2.0f*Q.k*Q.k);
  d[1]=2.0f*Q.i*Q.j+2.0f*Q.k*Q.r;
  d[2]=2.0f*Q.i*Q.k-2.0f*Q.j*Q.r;
  d[3]=2.0f*Q.i*Q.j-2.0f*Q.k*Q.r;
  d[4]=1.0f-(2.0f*Q.i*Q.i+2.0f*Q.k*Q.k);
  d[5]=2.0f*Q.j*Q.k+2.0f*Q.i*Q.r;
  d[6]=2.0f*Q.i*Q.k+2.0f*Q.j*Q.r;
  d[7]=2.0f*Q.j*Q.k-2.0f*Q.i*Q.r;
  d[8]=1.0f-(2.0f*Q.i*Q.i+2.0f*Q.j*Q.j);
 }
 static mat3x3 linear_interpolate(const mat3x3 &A,const mat3x3 &B,float P);
 // ortogonalna matica - 3 vektory su navzajom kolme a normalizovane
 void orthonormalize() {
  vec3 u1(operator() (0,0),operator() (1,0),operator() (2,0));
  vec3 u2(operator() (0,1),operator() (1,1),operator() (2,1));
  vec3 u3(operator() (0,2),operator() (1,2),operator() (2,2));

  vec3 p;
  vec3 w1=u1.normalize();
  vec3 w2=(u2-p.proj(w1,u2)); w2.normalize();
  vec3 w3=(u3-p.proj(w1,u3)-p.proj(w2,u3)); w3.normalize();

  operator() (0,0)=w1[0]; operator() (1,0)=w1[1]; operator() (2,0)=w1[2];
  operator() (0,1)=w2[0]; operator() (1,1)=w2[1]; operator() (2,1)=w2[2];
  operator() (0,2)=w3[0]; operator() (1,2)=w3[1]; operator() (2,2)=w3[2];

  if (!sensible()) {
   *this=mat3x3(vec3(1.0f,0.0f,0.0f),vec3(0.0f,1.0f,0.0f),vec3(0.0f,0.0f,1.0f));
  }
 }
 const float *get_d() { return(d); }
 const float *get_d() const { return(d); }
  void set_d(const float *D) {
  memcpy(d,D,9*sizeof(float));
 }
 vec3 get_col(unsigned int I) const {
  const unsigned int o=I*3;
  return(vec3(d[o],d[o+1],d[o+2]));
 }
 void set_col(unsigned int I,const vec3 &C) {
  const unsigned int o=I*3;
  d[o]  =C[0];
  d[o+1]=C[1];
  d[o+2]=C[2];
 }
 friend mat3x3 transpose(const mat3x3 &D);
  float trace(const mat3x3 &D) {
  return(D(0,0)+D(1,1)+D(2,2));
 }
 void show(const char *str="") const {
  loopi(3) {
   loopj(3) {
   }
  }
 }
 mat3x3 m3alpha(float A) {
  mat3x3 result(0);
  float s=(float) sin_deg(A);
  float c=(float) cos_deg(A);

  result(0,0)=1;
  result(1,1)=c;
  result(2,2)=c;
  result(2,1)=s;
  result(1,2)=-s;

  return(result);
 }
 mat3x3 m3beta(float B) {
  mat3x3 result(0);
  float s=(float) sin_deg(B);
  float c=(float) cos_deg(B);

  result(1,1)=1;
  result(2,2)=c;
  result(0,0)=c;
  result(0,2)=s;
  result(2,0)=-s;

  return(result);
 }
 mat3x3 m3gamma(float G) {
  mat3x3 result(0);
  float s=(float) sin_deg(G);
  float c=(float) cos_deg(G);

  result(2,2)=1;
  result(0,0)=c;
  result(1,1)=c;
  result(1,0)=s;
  result(0,1)=-s;

  return(result);
 }
};
// globalne operatory a funkcie suvisiace s triedou mat3x3 *****************
inline mat3x3 operator*(const mat3x3 &A,const float B) {
 mat3x3 result;
 loopi(9) result.d[i]=A.d[i]*B;
 return(result);
}
inline mat3x3 operator*(const float A,const mat3x3 &B) { return(B*A); }
inline mat3x3 operator*(const mat3x3 &A,const mat3x3 &B) {
 return(mat3x3(A.d[0]*B.d[0]+A.d[1]*B.d[3]+A.d[2]*B.d[6],
               A.d[0]*B.d[1]+A.d[1]*B.d[4]+A.d[2]*B.d[7],
               A.d[0]*B.d[2]+A.d[1]*B.d[5]+A.d[2]*B.d[8],
               A.d[3]*B.d[0]+A.d[4]*B.d[3]+A.d[5]*B.d[6],
               A.d[3]*B.d[1]+A.d[4]*B.d[4]+A.d[5]*B.d[7],
               A.d[3]*B.d[2]+A.d[4]*B.d[5]+A.d[5]*B.d[8],
               A.d[6]*B.d[0]+A.d[7]*B.d[3]+A.d[8]*B.d[6],
               A.d[6]*B.d[1]+A.d[7]*B.d[4]+A.d[8]*B.d[7],
               A.d[6]*B.d[2]+A.d[7]*B.d[5]+A.d[8]*B.d[8]));
}
inline mat3x3 transpose(const mat3x3 &D) {
 return(mat3x3(D(0,0),D(0,1),D(0,2),
               D(1,0),D(1,1),D(1,2),
               D(2,0),D(2,1),D(2,2)));
}
inline const mat3x3 &mat3x3_identity() {
 static const mat3x3 result(1.0f,0.0f,0.0f,
                            0.0f,1.0f,0.0f,
                            0.0f,0.0f,1.0f);
 return(result);
}
inline mat3x3 rotation_matrix(float A,const vec3 &D) {
 mat3x3 uut(D[0]*D[0],D[1]*D[0],D[2]*D[0],
            D[0]*D[1],D[1]*D[1],D[2]*D[1],
            D[0]*D[2],D[1]*D[2],D[2]*D[2]);
 mat3x3 s(0,D[2],-D[1],
          -D[2],0,D[0],
           D[1],-D[0],0);
 return(uut+(float) cos_deg(A)*(mat3x3_identity()-uut)+(float) sin_deg(A)*s);
}

// 12-prvkova matica *******************************************************
class mat4x3 {
 public:
  float d[16];
  // konstruktor
  mat4x3() {
   d[1]=d[2]=d[3]=d[4]=d[6]=d[7]=d[8]=d[9]=d[11]=0.0f;
   d[0]=d[5]=d[10]=1.0f;
   d[12]=d[13]=d[14]=0; d[15]=1.0f;
  }
  void set_diagonal(float A,float B,float C) { d[0]=A; d[5]=B; d[10]=C; }
  // nasobenie matic
  mat4x3 operator*(const mat4x3 M) const {
   mat4x3 result;
   result.d[0] = (M.d[0]*d[0]) + (M.d[4]*d[1]) + (M.d[8]*d[2]);
   result.d[4] = (M.d[0]*d[4]) + (M.d[4]*d[5]) + (M.d[8]*d[6]);
   result.d[8] = (M.d[0]*d[8]) + (M.d[4]*d[9]) + (M.d[8]*d[10]);
   result.d[1] = (M.d[1]*d[0]) + (M.d[5]*d[1]) + (M.d[9]*d[2]);
   result.d[5] = (M.d[1]*d[4]) + (M.d[5]*d[5]) + (M.d[9]*d[6]);
   result.d[9] = (M.d[1]*d[8]) + (M.d[5]*d[9]) + (M.d[9]*d[10]);
   result.d[2] = (M.d[2]*d[0]) + (M.d[6]*d[1]) + (M.d[10]*d[2]);
   result.d[6] = (M.d[2]*d[4]) + (M.d[6]*d[5]) + (M.d[10]*d[6]);
   result.d[10]= (M.d[2]*d[8]) + (M.d[6]*d[9]) + (M.d[10]*d[10]);
   result.d[3] = (M.d[3]*d[0]) + (M.d[7]*d[1]) + (M.d[11]*d[2])  + d[3];
   result.d[7] = (M.d[3]*d[4]) + (M.d[7]*d[5]) + (M.d[11]*d[6])  + d[7];
   result.d[11]= (M.d[3]*d[8]) + (M.d[7]*d[9]) + (M.d[11]*d[10]) + d[11];
   return(result);
  }
  // nasobenie matice vektorom
  vec3 operator*(const vec3 V) const {
   return(vec3(d[0]*V.x+d[1]*V.y+d[2]*V.z +d[3],
               d[4]*V.x+d[5]*V.y+d[6]*V.z +d[7],
               d[8]*V.x+d[9]*V.y+d[10]*V.z+d[11]));
  }
  vec3 mul(const vec3 &V) const {
   return((*this)*V);
  }
  // determinant matice
  float get_determinant() const;
  // inverzna matica
  void set_inverse(const mat4x3 &M);
  mat4x3 inverse() const {
   mat4x3 result;
   result.set_inverse(*this);
   return(result);
  }
  void invert() { set_inverse(*this); }

  vec3 mul_transp(const vec3 V) const {
   vec3 tmp=V;
   tmp.x-=d[3]; tmp.y-=d[7]; tmp.z-=d[11];
   return(vec3(d[0]*tmp.x+d[4]*tmp.y+d[8]*tmp.z,
               d[1]*tmp.x+d[5]*tmp.y+d[9]*tmp.z,
               d[2]*tmp.x+d[6]*tmp.y+d[10]*tmp.z));
  }

  vec3 get_column_vector(int I) const { return(vec3(d[I],d[I+4],d[I+8])); }

  void set_orientation_and_pos(const CQuaternion &Q,const vec3 &P) {
   d[0] = 1.0f-(2.0f*Q.j*Q.j+2.0f*Q.k*Q.k);
   d[1] = 2.0f*Q.i*Q.j+2.0f*Q.k*Q.r;
   d[2] = 2.0f*Q.i*Q.k-2.0f*Q.j*Q.r;
   d[3] = P.x;

   d[4] = 2.0f*Q.i*Q.j-2.0f*Q.k*Q.r;
   d[5] = 1.0f-(2.0f*Q.i*Q.i+2.0f*Q.k*Q.k);
   d[6] = 2.0f*Q.j*Q.k+2.0f*Q.i*Q.r;
   d[7] = P.y;

   d[8] = 2.0f*Q.i*Q.k+2.0f*Q.j*Q.r;
   d[9] = 2.0f*Q.j*Q.k-2.0f*Q.i*Q.r;
   d[10]= 1.0f-(2.0f*Q.i*Q.i+2.0f*Q.j*Q.j);
   d[11]= P.z;
  }
  void fillGLarray(float A[16]) const {
   A[0] = d[0]; A[1] = d[4]; A[2] = d[8];  A[3] = d[12];
   A[4] = d[1]; A[5] = d[5]; A[6] = d[9];  A[7] = d[13];
   A[8] = d[2]; A[9] = d[6]; A[10]= d[10]; A[11]= d[14];
   A[12]= d[3]; A[13]= d[7]; A[14]= d[11]; A[15]= d[15];
  }
};

// 16-prvkova matica *******************************************************
class mat16 {
 public:

  // udaje -----------------------------------------------------------------
  float m[16];

  // konstruktory ----------------------------------------------------------
  mat16() {
   m[0]=1.0f;  m[1]=0.0f;  m[2] =0.0f; m[3]=0.0f;
   m[4]=0.0f;  m[5]=1.0f;  m[6] =0.0f; m[7]=0.0f;
   m[8]=0.0f;  m[9]=0.0f;  m[10]=1.0f; m[11]=0.0f;
   m[12]=0.0f; m[13]=0.0f; m[14]=0.0f; m[15]=1.0f;
  }
  mat16(const mat16 &M) {
   (*this)=M;
  }

  // operatory -------------------------------------------------------------
  // index
  float &operator[](const int I) { return m[I]; }
  // nasobenie matic, vysledok sa ulozi do povodnej matice
  const mat16 operator*=(const mat16 M) {
   mat16 mat;

   mat.m[0] =m[0]*M.m[0] +m[4]*M.m[1] +m[8] *M.m[2] +m[12]*M.m[3];
   mat.m[1] =m[1]*M.m[0] +m[5]*M.m[1] +m[9] *M.m[2] +m[13]*M.m[3];
   mat.m[2] =m[2]*M.m[0] +m[6]*M.m[1] +m[10]*M.m[2] +m[14]*M.m[3];
   mat.m[3] =m[3]*M.m[0] +m[7]*M.m[1] +m[11]*M.m[2] +m[15]*M.m[3];

   mat.m[4] =m[0]*M.m[4] +m[4]*M.m[5] +m[8] *M.m[6] +m[12]*M.m[7];
   mat.m[5] =m[1]*M.m[4] +m[5]*M.m[5] +m[9] *M.m[6] +m[13]*M.m[7];
   mat.m[6] =m[2]*M.m[4] +m[6]*M.m[5] +m[10]*M.m[6] +m[14]*M.m[7];
   mat.m[7] =m[3]*M.m[4] +m[7]*M.m[5] +m[11]*M.m[6] +m[15]*M.m[7];

   mat.m[8] =m[0]*M.m[8] +m[4]*M.m[9] +m[8] *M.m[10]+m[12]*M.m[11];
   mat.m[9] =m[1]*M.m[8] +m[5]*M.m[9] +m[9] *M.m[10]+m[13]*M.m[11];
   mat.m[10]=m[2]*M.m[8] +m[6]*M.m[9] +m[10]*M.m[10]+m[14]*M.m[11];
   mat.m[11]=m[3]*M.m[8] +m[7]*M.m[9] +m[11]*M.m[10]+m[15]*M.m[11];

   mat.m[12]=m[0]*M.m[12]+m[4]*M.m[13]+m[8] *M.m[14]+m[12]*M.m[15];
   mat.m[13]=m[1]*M.m[12]+m[5]*M.m[13]+m[9] *M.m[14]+m[13]*M.m[15];
   mat.m[14]=m[2]*M.m[12]+m[6]*M.m[13]+m[10]*M.m[14]+m[14]*M.m[15];
   mat.m[15]=m[3]*M.m[12]+m[7]*M.m[13]+m[11]*M.m[14]+m[15]*M.m[15];

   *this=mat;

   return(*this);
  }
  // nasobenie matic, vysledok sa ulozi do novej matice
  const mat16 operator*(const mat16 M) {
   mat16 mat;

   mat.m[0] =m[0]*M.m[0] +m[4]*M.m[1] +m[8] *M.m[2] +m[12]*M.m[3];
   mat.m[1] =m[1]*M.m[0] +m[5]*M.m[1] +m[9] *M.m[2] +m[13]*M.m[3];
   mat.m[2] =m[2]*M.m[0] +m[6]*M.m[1] +m[10]*M.m[2] +m[14]*M.m[3];
   mat.m[3] =m[3]*M.m[0] +m[7]*M.m[1] +m[11]*M.m[2] +m[15]*M.m[3];

   mat.m[4] =m[0]*M.m[4] +m[4]*M.m[5] +m[8] *M.m[6] +m[12]*M.m[7];
   mat.m[5] =m[1]*M.m[4] +m[5]*M.m[5] +m[9] *M.m[6] +m[13]*M.m[7];
   mat.m[6] =m[2]*M.m[4] +m[6]*M.m[5] +m[10]*M.m[6] +m[14]*M.m[7];
   mat.m[7] =m[3]*M.m[4] +m[7]*M.m[5] +m[11]*M.m[6] +m[15]*M.m[7];

   mat.m[8] =m[0]*M.m[8] +m[4]*M.m[9] +m[8] *M.m[10]+m[12]*M.m[11];
   mat.m[9] =m[1]*M.m[8] +m[5]*M.m[9] +m[9] *M.m[10]+m[13]*M.m[11];
   mat.m[10]=m[2]*M.m[8] +m[6]*M.m[9] +m[10]*M.m[10]+m[14]*M.m[11];
   mat.m[11]=m[3]*M.m[8] +m[7]*M.m[9] +m[11]*M.m[10]+m[15]*M.m[11];

   mat.m[12]=m[0]*M.m[12]+m[4]*M.m[13]+m[8] *M.m[14]+m[12]*M.m[15];
   mat.m[13]=m[1]*M.m[12]+m[5]*M.m[13]+m[9] *M.m[14]+m[13]*M.m[15];
   mat.m[14]=m[2]*M.m[12]+m[6]*M.m[13]+m[10]*M.m[14]+m[14]*M.m[15];
   mat.m[15]=m[3]*M.m[12]+m[7]*M.m[13]+m[11]*M.m[14]+m[15]*M.m[15];

   return(mat);
  }
  // nasobenie vektora a matice
  const vec3 operator*(const vec3 V) {
   float a,b,c;
   a=V.x*m[0]+V.y*m[4]+V.z*m[8] +m[12];
   b=V.x*m[1]+V.y*m[5]+V.z*m[9] +m[13];
   c=V.x*m[2]+V.y*m[6]+V.z*m[10]+m[14];
   return(vec3(a,b,c));
  }

  // funkcie ---------------------------------------------------------------
  // priradenie prvkov matici
  void set(const float M[16]) {
   m[0]=M[0];   m[1]=M[1];   m[2] =M[2];  m[3]=M[3];
   m[4]=M[4];   m[5]=M[5];   m[6] =M[6];  m[7]=M[7];
   m[8]=M[8];   m[9]=M[9];   m[10]=M[10]; m[11]=M[11];
   m[12]=M[12]; m[13]=M[13]; m[14]=M[14]; m[15]=M[15];
  }
  // ziskanie prvkov matice
  float *get() {
   return(m);
  }
  // jednotkova matica
  void identity() {
   m[0]=1.0f;  m[1]=0.0f;  m[2] =0.0f; m[3]=0.0f;
   m[4]=0.0f;  m[5]=1.0f;  m[6] =0.0f; m[7]=0.0f;
   m[8]=0.0f;  m[9]=0.0f;  m[10]=1.0f; m[11]=0.0f;
   m[12]=0.0f; m[13]=0.0f; m[14]=0.0f; m[15]=1.0f;
  }
  // transponovana matica
  void transpose(const mat16 M) {
   m[0]=M.m[0]; m[4]=M.m[1]; m[8]= M.m[2];
   m[1]=M.m[4]; m[5]=M.m[5]; m[9]= M.m[6];
   m[2]=M.m[8]; m[6]=M.m[9]; m[10]=M.m[10];
  }
  // posun
  void translate(float X,float Y,float Z) {
   m[3]+=X; m[7]+=Y; m[11]+=Z;
  }
  // rotacia
  void rotate(float X,float Y,float Z) {
   float cx,cy,cz,sx,sy,sz;
   mat16 mat;

   cx=cosf(X); sx=sinf(X);
   cy=cosf(Y); sy=sinf(Y);
   cz=cosf(Z); sz=sinf(Z);

   mat.m[0] =cy*cz;
   mat.m[1] =cy*sz;
   mat.m[2] =-sy;
   mat.m[4] =(sx*sy*cz)-(cx*sz);
   mat.m[5] =(sx*sy*sz)+(cx*cz);
   mat.m[6] =sx*cy;
   mat.m[8] =(cx*sy*cz)+(sx*sz);
   mat.m[9] =(cx*sy*sz)-(sx*cz);
   mat.m[10]=cx*cy;

   (*this)=(*this)*mat;
  }

  float det() const {
   float det;
   det =m[0]*m[5]*m[10];
   det+=m[4]*m[9]*m[2];
   det+=m[8]*m[1]*m[6];
   det-=m[8]*m[5]*m[2];
   det-=m[4]*m[1]*m[10];
   det-=m[0]*m[9]*m[6];

   return(det);
  }

  mat16 inverse() const {
   mat16 ret;
   float idet=1.0f/det();
   ret[0] = (m[5]*m[10]-m[9]*m[6])*idet;
   ret[1] =-(m[1]*m[10]-m[9]*m[2])*idet;
   ret[2] = (m[1]*m[6] -m[5]*m[2])*idet;
   ret[3] = 0.0;
   ret[4] =-(m[4]*m[10]-m[8]*m[6])*idet;
   ret[5] = (m[0]*m[10]-m[8]*m[2])*idet;
   ret[6] =-(m[0]*m[6] -m[4]*m[2])*idet;
   ret[7] = 0.0;
   ret[8] = (m[4]*m[9]-m[8]*m[5])*idet;
   ret[9] =-(m[0]*m[9]-m[8]*m[1])*idet;
   ret[10]= (m[0]*m[5]-m[4]*m[1])*idet;
   ret[11]= 0.0;
   ret[12]=-(m[12]*ret[0]+m[13]*ret[4]+m[14]*ret[8]);
   ret[13]=-(m[12]*ret[1]+m[13]*ret[5]+m[14]*ret[9]);
   ret[14]=-(m[12]*ret[2]+m[13]*ret[6]+m[14]*ret[10]);
   ret[15]= 1.0;

   return(ret);
  }
};

struct triangle
{
 // konstruktory
 triangle() {}
 triangle(const vec3 &A,const vec3 &B,const vec3 &C,
          const vec2 T0,const vec2 T1,const vec2 T2) {
  v0=A; v1=B; v2=C;              // vrcholy
  t0=T0; t1=T1; t2=T2;           // texturovacie suradnice
  e0=v1-v0; e1=v2-v0;            // 2 strany
  n=cross(e0,e1); n.normalize(); // normala
  // vypocet prvku D rovnice roviny, ktoru tvori trojuholnik
  // D=-(AX+BY+CZ)
  // A,B,C - normala trojuholnika
  // X,Y,Z - lubovolny vrchol trojuholnika
  d=-(n.x*v0.x+n.y*v0.y+n.z*v0.z);
  // vypocet koliznych normal
  p=v1-v0; q=cross(p,n); q.normalize();
  cn[0].x=q.x; cn[0].y=q.y; cn[0].z=q.z;
  cn[0].w=-(q.x*v0.x+q.y*v0.y+q.z*v0.z);
  p=v2-v1; q=cross(p,n); q.normalize();
  cn[1].x=q.x; cn[1].y=q.y; cn[1].z=q.z;
  cn[1].w=-(q.x*v1.x+q.y*v1.y+q.z*v1.z);
  p=v0-v2; q=cross(p,n); q.normalize();
  cn[2].x=q.x; cn[2].y=q.y; cn[2].z=q.z;
  cn[2].w=-(q.x*v2.x+q.y*v2.y+q.z*v2.z);
 }

 vec3 v0,v1,v2; // vrcholy trojuholnika
 vec2 t0,t1,t2; // texturovacie suradnice
 vec3 e0,e1;    // 2 strany zacinajuce vo vrchole v0
 vec3 n;        // normala trojuholnika

 float d; // prvok D rovnice roviny, ktoru tvori trojuholnik,
          // prvkami A, B, C su prvky normaly trojuholnika
 vec4 cn[3]; // prvky A, B, C "koliznych normal" a prvok D,
             // ktore spolu definuju rovnice rovin pre jednotlive
             // strany trojuholnika - vyuzivaju sa pre urcenie, ci
             // dany 3D bod patri/nepatri trojuholniku
 vec3 p,q; // pomocne vektory
};

struct ellipsoid
{
 ellipsoid(const vec3 &A):s(A) {}
 vec3 s; // rozmery elipsoidu
};

// pomerna interpolacia 8 hodnot
// X,Y,Z su v rozsahu 0.0 - 1.0
// ak X,Y,Z = 0.5, interpolacia = aritmeticky priemer 8-mich hodnot
template<class T>
T interp3D(const T &V000,const T &V001,
           const T &V010,const T &V011,
           const T &V100,const T &V101,
           const T &V110,const T &V111,
           float X,float Y,float Z) // pomer interpolovanych hodnot
{
 T a=(1.0f-Y)*(V000*(1.0f-X)+V100*X)+Y*(V010*(1.0f-X)+V110*X);
 T b=(1.0f-Y)*(V001*(1.0f-X)+V101*X)+Y*(V011*(1.0f-X)+V111*X);
 return((1.0f-Z)*a+Z*b);
}

// trieda premennych a funkcii (objektov) pre pracu s matematikou
class CMath3d {
public:
 CMath3d();  // konstruktor
 ~CMath3d(); // destruktor

 bool was_clip;

 // kolizia svetelneho luca a trojuholnika
 bool ray_tri_collision(vec3 P,vec3 L,vec3 N,float D,vec4 CN0,vec4 CN1,vec4 CN2);
 // priesecnik vektora a roviny
 bool vec_plane_intersection(vec3 A,vec3 B,vec3 N,float D);
 // ziskanie priesecnika luca a trojuholnika
 vec3 get_ipoint(void);
 // test polohy bodu a trojuholnika
 bool point_in_triangle(vec3 P,vec4 CN0,vec4 CN1,vec4 CN2);
 // kolizia bodu so stranami trojuholnika
 bool edge_collision(vec3 *P,vec3 *S,vec3 A,vec3 B,vec3 C,float R);
 // vypocet polohy bodu na ciare A-B, ktory je najblizsie k bodu C
 vec3 point_on_line(vec3 A,vec3 B,vec3 C);

 // orezanie stran 4-uholnika podla stran 3-uholnika
 int quad_tri_clip(vec3 *V,vec2 *T,vec4 *TRI);
 // orezanie stran 3-uholnika podla stran 3D hranola
 int tri_box_clip(vec3 *V,vec4 *P);
 // test ci trojuholnik zasahuje do 3D hranola
 bool tri_in_box(vec3 A,vec3 B,vec3 C,vec4 *P);
private:
 vec3 v0,v1,v2; // pomocne vektory
 vec2 v3;       // pomocny vektor

 vec3 p[9]; // vrcholy n-uholnika
 vec2 q[9]; // texturovacie suradnice n-uholnika
 int vnum;  // aktualny pocet vrcholov n-uholnika

 float a,b,c; // pomocne premenne
 int ai;      // pomocna premena
};

#endif
